// ======================================================================
//
// gl_d3d_5.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "first.h"
#include "Gl_d3d_5.h"

#include "Geometry.h"
#include "Gl_init.h"
#include "Gl_sfx.h"
#include "Material.h"
#include "MaterialData_d3d_5.h"
#include "MemoryBlockManager.h"
#include "PaletteData.h"
#include "PixelIndex.h"
#include "TranslucentQueue.h"
#include "uvstruct.h"

#include <stdlib.h>
#include <float.h>

// ======================================================================

int                             GL_d3d_5::matR;
int                             GL_d3d_5::matG;
int                             GL_d3d_5::matB;

int                             GL_d3d_5::tripleBuffer;
const Direct3d::Driver         *GL_d3d_5::driver;

LPDIRECT3DDEVICE2               GL_d3d_5::device;
LPDIRECT3DVIEWPORT2             GL_d3d_5::viewport;
LPDIRECT3DMATERIAL2             GL_d3d_5::background;
D3DTLVERTEX                     GL_d3d_5::vertexArray[Geometry::TVERTEX_BUFFER];

bool                            GL_d3d_5::bARGB_0555;
bool                            GL_d3d_5::bARGB_1555;
bool                            GL_d3d_5::bARGB_4444;

bool                            GL_d3d_5::materialTranslucent;
PrimitiveProcessing             GL_d3d_5::primitiveProcessing;

// ======================================================================

GL_d3d_5::GL_d3d_5(void)
{
	tripleBuffer = 0;
	driver       = NULL;
	device       = NULL;
	viewport     = NULL;
	background   = NULL;

	submitPointListFunction     = passPointList;
	submitLineListFunction      = passLineList;
	submitLineStripFunction     = passLineStrip;
	submitTriangleListFunction  = passTriangleList;
	submitTriangleStripFunction = passTriangleStrip;
	submitTriangleFanFunction   = passTriangleFan;

	TranslucentQueue::install(this, passPointList, passLineList, passTriangleList);

	primitiveProcessing = PP_PassAll;
	materialTranslucent = false;
}

// ----------------------------------------------------------------------

GL_d3d_5::~GL_d3d_5(void)
{
	HRESULT result;

	MaterialData_d3d_5::remove();

	TranslucentQueue::remove();

	if (viewport)
	{
		result = viewport->Release();
		viewport= NULL;
	}

	if (device)
	{
		result = device->Release();
		device = NULL;
	}

	driver = NULL;
	Direct3d::remove();
	Geometry::remove();
}

// ----------------------------------------------------------------------

void GL_d3d_5::getInformation(Information *information)
{
	GL_abstract::getInformation(information);

	information->name         = "Direct3d";
	information->versionMinor = 0;
	information->ranking      = -100;
}

// ----------------------------------------------------------------------

bool GL_d3d_5::query(int minor)
{
	return pickDriver(minor) != NULL;
}

// ----------------------------------------------------------------------

const Direct3d::Driver *GL_d3d_5::pickDriver(int minor)
{
	int                     i, dRating, bestRating;
	const Direct3d::Driver *d, *best;

	// enumerate all available 3d devices
	Direct3d::enumerate();

	// handle specifically requested driver
	if (minor)
		return Direct3d::getDriver(minor);

	// check out all the drivers looking for the best one
	for (best = NULL, bestRating = 0, i = 1; (d = Direct3d::getDriver(i)) != NULL; ++i)
		if (d->isHardware && d->canTexture && d->zBuffer)
		{
			dRating = (d->rgbLighting ? 8 : 0) | (d->alphaBlending ? 4 : 0) | (d->bilinearFiltering ? 2 : 0);

			if (!best || dRating > bestRating)
			{
				best = d;
				bestRating = dRating;
			}
		}

	return best;
}

// ----------------------------------------------------------------------
 
bool GL_d3d_5::init(const GL_input *input, GL_output *output)
{
	HRESULT              result;
  D3DMATERIAL          mat;
  D3DMATERIALHANDLE    matHandle;
	DDSCAPS              ddscaps;
	DWORD                total, avail;

	driver = pickDriver(input->minor);
	DEBUG_FATAL(!driver, ("gl_d3d_5::init no driver"));

	// install the DD layer
	if (!GL_dd::init(input, output, driver->directDrawDriver))
		return false;

	// install D3D
	Direct3d::install(driver);

	// set double or triple buffering
	DirectDraw::doubleBuffer();

	// create the rendering surface
	GL_dd::createSurfaces(input, output);

	// create zbuffer
	createZBuffer(DirectDraw::getBackSurface(), width, height);

	// check b2DDuring 3d
	DDCAPS DDHWCaps, DDSWCaps;
	memset(&DDHWCaps, 0, sizeof(DDCAPS));
	DDHWCaps.dwSize = sizeof(DDCAPS);
	memset(&DDSWCaps, 0, sizeof(DDCAPS));
	DDSWCaps.dwSize = sizeof(DDCAPS);

	DirectDraw::getDirectDrawObject()->GetCaps(&DDHWCaps, &DDSWCaps);

	// create the direct3d device
	result = Direct3d::d3d->CreateDevice(driver->guid, DirectDraw::getBackSurface(), &device);
	FATAL_HR("gl_d3d_5::init createDev failed %d", result);

	// create viewport
	result = Direct3d::d3d->CreateViewport(&viewport, NULL);
	FATAL_HR("gl_d3d_5::init createvp %d", result);

	// bind the viewport into the device
	result = device->AddViewport(viewport);
	FATAL_HR("gl_d3d_5::init addvp %d", result);

	setViewport(0, 0, width-1, height-1);

	// create the background material
	memset(&mat, 0, sizeof(mat));
	mat.dwSize = sizeof(mat);
	mat.dwRampSize = 1;
	result = Direct3d::d3d->CreateMaterial(&background, NULL);
	FATAL_HR("gl_d3d_5::init createmat %d", result);
	result = background->SetMaterial(&mat);
	FATAL_HR("gl_d3d_5::init setmat %d", result);
	result = background->GetHandle(device, &matHandle);
	FATAL_HR("gl_d3d_5::init geth %d", result);
	result = viewport->SetBackground(matHandle);
	FATAL_HR("gl_d3d_5::init setb %d", result);

	// tell the device what viewport to use
	result = device->SetCurrentViewport(viewport);
	FATAL_HR("gl_d3d_5::init setvp %d", result);

	// find out how much texture memory is available
	ddscaps.dwCaps = DDSCAPS_TEXTURE;
	result = DirectDraw::getDirectDrawObject()->GetAvailableVidMem(&ddscaps, &total, &avail);
	FATAL_HR("gl_d3d_5::init getAvailVidMem %d", result);
	MESSAGE(true, ("gl_d3d_5::init texture %d=total %d=avail\n", total, avail));

	getCaps(device);

	// figure out the best texture destination for the movie player
	if (bARGB_0555 || bARGB_1555)
	{
		output->rMovieSizeShift   = 3;
		output->rMoviePlaceShift  = 10;
		output->gMovieSizeShift   = 3;
		output->gMoviePlaceShift  = 5;
		output->bMovieSizeShift   = 3;
		output->bMoviePlaceShift  = 0;
	}
	else
		if (bARGB_4444)
		{
			output->rMovieSizeShift   = 4;
			output->rMoviePlaceShift  = 8;
			output->gMovieSizeShift   = 4;
			output->gMoviePlaceShift  = 4;
			output->bMovieSizeShift   = 4;
			output->bMoviePlaceShift  = 0;
		}

	// setup some basic state info for the device
	result = device->SetRenderState(D3DRENDERSTATE_SHADEMODE, D3DSHADE_GOURAUD);
	FATAL_HR("gl_d3d_5::init shademode %d", result);
	result = device->SetRenderState(D3DRENDERSTATE_ZENABLE, TRUE);
	FATAL_HR("gl_d3d_5::init zenable %d", result);
	result = device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE, TRUE);
	FATAL_HR("gl_d3d_5::init zwriteenable %d", result);
	result = device->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_LESSEQUAL);
	FATAL_HR("gl_d3d_5::init zfunc %d", result);
	result = device->SetRenderState(D3DRENDERSTATE_SPECULARENABLE, FALSE);
	FATAL_HR("gl_d3d_5::init specularenable %d", result);
	result = device->SetRenderState(D3DRENDERSTATE_COLORKEYENABLE, TRUE);
	FATAL_HR("gl_d3d_5::init colorkeyenable %d", result);
	result = device->SetRenderState(D3DRENDERSTATE_TEXTUREPERSPECTIVE, TRUE);
	FATAL_HR("gl_d3d_5::init texturepersp %d", result);

	// enable bilinear filtering if we can
	if (driver->bilinearFiltering)
	{
		result = device->SetRenderState(D3DRENDERSTATE_TEXTUREMAG, D3DFILTER_LINEAR);
		FATAL_HR("gl_d3d_5::init texture mag %d", result);
		result = device->SetRenderState(D3DRENDERSTATE_TEXTUREMIN, D3DFILTER_LINEAR);
		FATAL_HR("gl_d3d_5::init texture min %d", result);
	}

	// enable dithering if supported
	if (driver->dithering)
	{
		result = device->SetRenderState(D3DRENDERSTATE_DITHERENABLE, TRUE);
		FATAL_HR("gl_d3d_5::init dither %d", result);
	}

	// setup alpha blending modes
	if (driver->alphaBlending && bARGB_4444)
	{
		result = device->SetRenderState(D3DRENDERSTATE_SRCBLEND, D3DBLEND_SRCALPHA);
		FATAL_HR("gl_d3d_5::init srcBlend %d", result);
		result = device->SetRenderState(D3DRENDERSTATE_DESTBLEND, D3DBLEND_INVSRCALPHA);
		FATAL_HR("gl_d3d_5::init destBlend %d", result);
	}

	// setup the geometry engine
	Geometry::install(this);

  // setup return values to the game
	output->lightingRGB = driver->rgbLighting;
	output->translucency = driver->alphaBlending && bARGB_4444;
	output->bilinearFiltering = driver->bilinearFiltering;
	output->textureMemory = avail;

	// make sure we have a 555 texture format available
	if (!bARGB_0555 && !bARGB_1555)
		return false;

	return true;
}

// ----------------------------------------------------------------------

HRESULT WINAPI GL_d3d_5::enumTextureFormatsCallback(LPDDSURFACEDESC pddsd, LPVOID /* lpContext */)
{
	DPRINT_LOG(("gl_d3d_5::EnumTexture %dx%dx%-2d 0x%x=f 0x%x=cc 0x%x=r 0x%x=g 0x%x=b 0x%x=a\n", 
		pddsd->dwWidth, pddsd->dwHeight,
		pddsd->ddpfPixelFormat.dwRGBBitCount,
		pddsd->ddpfPixelFormat.dwFlags,
		pddsd->ddpfPixelFormat.dwFourCC,
		pddsd->ddpfPixelFormat.dwRBitMask,
		pddsd->ddpfPixelFormat.dwGBitMask,
		pddsd->ddpfPixelFormat.dwBBitMask,
		pddsd->ddpfPixelFormat.dwRGBAlphaBitMask));

	if (pddsd->ddpfPixelFormat.dwRGBBitCount == 16 && pddsd->ddpfPixelFormat.dwFlags & DDPF_RGB)
	{
		if ((pddsd->ddpfPixelFormat.dwRBitMask        == 0x7c00) &&
				(pddsd->ddpfPixelFormat.dwGBitMask        == 0x03e0) &&
				(pddsd->ddpfPixelFormat.dwBBitMask        == 0x001f) &&
				(pddsd->ddpfPixelFormat.dwRGBAlphaBitMask == 0x0000))
			bARGB_0555 = true;

		if ((pddsd->ddpfPixelFormat.dwRBitMask        == 0x7c00) &&
				(pddsd->ddpfPixelFormat.dwGBitMask        == 0x03e0) &&
				(pddsd->ddpfPixelFormat.dwBBitMask        == 0x001f) &&
				(pddsd->ddpfPixelFormat.dwRGBAlphaBitMask == 0x8000))
			bARGB_1555 = true;

		if ((pddsd->ddpfPixelFormat.dwRBitMask        == 0x0f00) &&
				(pddsd->ddpfPixelFormat.dwGBitMask        == 0x00f0) &&
				(pddsd->ddpfPixelFormat.dwBBitMask        == 0x000f) &&
				(pddsd->ddpfPixelFormat.dwRGBAlphaBitMask == 0xf000))
			bARGB_4444 = true;
	}

	return D3DENUMRET_OK;
}

// ----------------------------------------------------------------------

void GL_d3d_5::getCaps(LPDIRECT3DDEVICE2 device)
{
	HRESULT result;

	// check for texture formats
	bARGB_0555 = false;
	bARGB_1555 = false;
	bARGB_4444 = false;
	result = device->EnumTextureFormats(enumTextureFormatsCallback, NULL);
	FATAL_HR("gl_d3d_5::getCaps enumTxmForm %d", result);
}

// ----------------------------------------------------------------------

void GL_d3d_5::setActive(bool)
{
}

// ----------------------------------------------------------------------

void GL_d3d_5::setViewport(int x0, int y0, int x1, int y1)
{
	D3DVIEWPORT2         viewData;
	HRESULT              result;

	memset(&viewData, 0, sizeof(viewData));
	viewData.dwSize = sizeof(viewData);
	viewData.dwX          = x0;
	viewData.dwY          = y0;
	viewData.dwWidth      = x1 - x0 + 1;
	viewData.dwHeight     = y1 - y0 + 1;
	viewData.dvClipX      = -1;
	viewData.dvClipY      = -1;
	viewData.dvClipWidth  = 2;
	viewData.dvClipHeight = 2;
	viewData.dvMinZ       = 0;
	viewData.dvMaxZ       = 1;
	result = viewport->SetViewport2(&viewData);
	FATAL_HR("gl_d3d_5::setViewport setvp %d", result);
}

// ----------------------------------------------------------------------

bool GL_d3d_5::lock(void)
{
	D3DRECT d3dRect;
	HRESULT result;

	// check if there is a background color to clear to
	if (backgroundMaterial)
	{
		// get the constant RGB colors
		int r = backgroundMaterial->getR();
		int g = backgroundMaterial->getG();
		int b = backgroundMaterial->getB();

		// clear the viewport using the constant RGB colors
		DirectDraw::clear(DirectDraw::makeLFB(r, g, b));
	}

	// clear the zbuffer
	d3dRect.lX1 = 0;
	d3dRect.lY1 = 0;
	d3dRect.lX2 = width;
	d3dRect.lY2 = height;
	result = viewport->Clear(1, &d3dRect, D3DCLEAR_ZBUFFER);
	FATAL_HR("gl_d3d_5::lock clear %d", result);

	// start the scene
	result = device->BeginScene();
	FATAL_HR("gl_d3d_5::lock beginScene %d", result);

	return true;
}

// ----------------------------------------------------------------------

void GL_d3d_5::unlock(bool display)
{
	HRESULT               result;

	// turn off z-buffer writing for alpha blended polygons
	result = device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE, FALSE);
	FATAL_HR("gl_d3d_5::unlock zenable %d", result);

	// draw the translucent primitives
	TranslucentQueue::draw();

	// turn the z buffer writing back on
	result = device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE, TRUE);
	FATAL_HR("gl_d3d_5::unlock zenable %d", result);

	// end the scene
	result = device->EndScene();
	FATAL_HR("gl_d3d_5::unlock endScene %d", result);

	// end scene has the effect of flushing the pipeline
	MaterialData_d3d_5::flushBatch(true);

	// display the new frame
	GL_dd::unlock(display);
}

// ----------------------------------------------------------------------

void GL_d3d_5::setAlphaBlendEnabled(bool enabled)
{
	HRESULT result;

	result = device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, enabled);
	FATAL_HR("gl_d3d_5::setAlphaBlendEnabled %d", result);
}

// ----------------------------------------------------------------------

int GL_d3d_5::getD3DBlend(AlphaBlendFunction function)
{
	switch (function)
	{

		case ABF_Zero:                        return D3DBLEND_ZERO;    
		case ABF_One:                         return D3DBLEND_ONE;
		case ABF_SourceColor:                 return D3DBLEND_SRCCOLOR;          
		case ABF_InverseSourceColor:          return D3DBLEND_INVSRCCOLOR;              
		case ABF_SourceAlpha:                 return D3DBLEND_SRCALPHA;          
		case ABF_InverseSourceAlpha:          return D3DBLEND_INVSRCALPHA;              
		case ABF_DestinationAlpha:            return D3DBLEND_DESTALPHA;               
		case ABF_InverseDestinationAlpha:     return D3DBLEND_INVDESTALPHA;                   
		case ABF_DestinationColor:            return D3DBLEND_DESTCOLOR;               
		case ABF_InverseDestinationColor:     return D3DBLEND_INVDESTCOLOR;                   
		case ABF_SourceAlphaSaturate:         return D3DBLEND_SRCALPHASAT;               
	}

	DEBUG_FATAL(true, ("Gl_d3d_5::getD3DBlend bad parameter"));
	return 0;
}

// ----------------------------------------------------------------------

void GL_d3d_5::setAlphaBlendFunctions(AlphaBlendFunction source, AlphaBlendFunction destination)
{
	HRESULT result;

	result = device->SetRenderState(D3DRENDERSTATE_SRCBLEND, getD3DBlend(source));
	FATAL_HR("gl_d3d_5::setAlphaBlendFunctions srcBlend %d", result);

	result = device->SetRenderState(D3DRENDERSTATE_DESTBLEND, getD3DBlend(destination));
	FATAL_HR("gl_d3d_5::setAlphaBlendFunctions destBlend %d", result);
}

// ----------------------------------------------------------------------

void GL_d3d_5::setPrimitiveProcessing(PrimitiveProcessing newPrimitiveProcessing)
{
	enum
	{
		Pass,
		Queue,
		Test
	} result;

	primitiveProcessing = newPrimitiveProcessing;

	switch (primitiveProcessing)
	{
		case PP_PassAll:
			result = Pass;
			break;

		case PP_QueueAll:
			result = Queue;
			break;

		case PP_CheckMaterial:
			result = materialTranslucent ? Queue : Pass;
			break;

		case PP_CheckVertices:
			result = Test;
			break;

		case PP_CheckMaterialAndVertices:
			result = materialTranslucent ? Queue : Test;
			break;

		default:
			DEBUG_FATAL(true, ("Gl_d3d_5::setPrimitiveProcessing bad case"));
			result = Pass;
	}

	switch (result)
	{
		case Pass:
			submitPointListFunction     = passPointList;
			submitLineListFunction      = passLineList;
			submitLineStripFunction     = passLineStrip;
			submitTriangleListFunction  = passTriangleList;
			submitTriangleStripFunction = passTriangleStrip;
			submitTriangleFanFunction   = passTriangleFan;
			break;

		case Queue:
			submitPointListFunction     = queuePointList;
			submitLineListFunction      = queueLineList;
			submitLineStripFunction     = queueLineStrip;
			submitTriangleListFunction  = queueTriangleList;
			submitTriangleStripFunction = queueTriangleStrip;
			submitTriangleFanFunction   = queueTriangleFan;
			break;

		case Test:
			submitPointListFunction     = testPointList;
			submitLineListFunction      = testLineList;
			submitLineStripFunction     = testLineStrip;
			submitTriangleListFunction  = testTriangleList;
			submitTriangleStripFunction = testTriangleStrip;
			submitTriangleFanFunction   = testTriangleFan;
			break;
	}
}

// ----------------------------------------------------------------------

void GL_d3d_5::createZBuffer(LPDIRECTDRAWSURFACE surface, int width, int height)
{
	HRESULT             result;
	DDSURFACEDESC       sd;
	DWORD               devDepth;
	LPDIRECTDRAWSURFACE zSurface;

	// clear and init some fields of sd
	memset(&sd, 0, sizeof(sd));
	sd.dwSize         = sizeof(sd);
	sd.dwFlags        = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS | DDSD_ZBUFFERBITDEPTH;
	sd.ddsCaps.dwCaps = DDSCAPS_ZBUFFER;
	sd.dwWidth        = width;
	sd.dwHeight       = height;

	// hardware drivers must have their zbuffers in video ram
	if (driver->isHardware)
		sd.ddsCaps.dwCaps |= DDSCAPS_VIDEOMEMORY;
	else
		sd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;

	// choose the z buffer depth
	devDepth = driver->primary->dwDeviceZBufferBitDepth;
	if (devDepth & DDBD_16)
		sd.dwZBufferBitDepth = 16;
	else
		if (devDepth & DDBD_24)
			sd.dwZBufferBitDepth = 24;
		else
			if (devDepth & DDBD_32)
				sd.dwZBufferBitDepth = 32;
			else
				if (devDepth & DDBD_8)
					sd.dwZBufferBitDepth = 8;
				else
					FATAL(true, ("GL_d3d_5::createZBuffer bad depth"));

	DirectDraw::createSurface(&zSurface, &sd);
	
	// attach the Z-buffer to the back buffer
	result = surface->AddAttachedSurface(zSurface);
	FATAL_HR("GL_d3d_5::createZBuffer attach failed %d", result);
}

// ----------------------------------------------------------------------

PaletteData *GL_d3d_5::paletteCreateData(const Palette *pal)
{
	UNREF(pal);
	
	// return a non-null element so the palette thinks it has data loaded
	return NEW PaletteData();
}

// ----------------------------------------------------------------------

void GL_d3d_5::paletteDestroyData(PaletteData *palData)
{
	delete palData;
}

// ----------------------------------------------------------------------

MaterialData *GL_d3d_5::materialCreateData(const StaticMaterial *material)
{
	return NEW MaterialData_d3d_5(material);
}

// ----------------------------------------------------------------------

void GL_d3d_5::materialDestroyData(MaterialData *mat)
{
	delete mat;
}

// ----------------------------------------------------------------------

void GL_d3d_5::beginRender(const Camera &camera)
{
	int x0, y0, x1, y1;

	// tell the geometry engine about the new camera
	Geometry::setCamera(camera);

	// setup to clip to the viewport
	camera.getViewport(&x0, &y0, &x1, &y1);
	setViewport(x0, y0, x1, y1);
}

// ----------------------------------------------------------------------

void GL_d3d_5::endRender(void)
{
	// stop clipping to the camera's viewport
	setViewport(0, 0, width-1, height-1);
}

// ----------------------------------------------------------------------

void GL_d3d_5::setLightingEnabled(bool lightingEnabled)
{
	Geometry::setLightingEnabled(lightingEnabled);
}

// ----------------------------------------------------------------------

void GL_d3d_5::setLightList(Light **lightList, int lightCount)
{
	Geometry::setLightList(lightList, lightCount);
}

// ----------------------------------------------------------------------

bool GL_d3d_5::setVertexList(const Vector *vertices, int vertexCount, const Vector *normal, int normalCount)
{
	return Geometry::setVertexList(vertices, vertexCount, normal, normalCount);
}

// ----------------------------------------------------------------------

void GL_d3d_5::setObjectToWorld(const Xform &objectToWorld)
{
	Geometry::setObjectToWorld(objectToWorld);
}

// ----------------------------------------------------------------------

void GL_d3d_5::setSolid(void)
{
	HRESULT  result;

	// set no texture
	result = device->SetRenderState(D3DRENDERSTATE_TEXTUREHANDLE, NULL);
	FATAL_HR("gl_d3d_5::setSolid render txhandle %d", result);

	// set no alpha
	result = device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
	FATAL_HR("gl_d3d_5::setSolid alphaBlend %d", result);
}

// ----------------------------------------------------------------------

void GL_d3d_5::setMaterial(const StaticMaterial *material)
{
	MaterialData_d3d_5    *materialData;

	TranslucentQueue::setMaterial(material);

	// unset the material
	if (material == NULL)
		return;

	materialData = static_cast<MaterialData_d3d_5 *>(material->getData());
	materialTranslucent = materialData->hasTranslucency();

	if (primitiveProcessing == PP_CheckMaterial)
		if (materialTranslucent)
		{
			submitPointListFunction     = queuePointList;
			submitLineListFunction      = queueLineList;
			submitLineStripFunction     = queueLineStrip;
			submitTriangleListFunction  = queueTriangleList;
			submitTriangleStripFunction = queueTriangleStrip;
			submitTriangleFanFunction   = queueTriangleFan;

			// -qq- OPTIMIZE should not have to setup the material in this case
		}
		else
		{
			submitPointListFunction     = passPointList;
			submitLineListFunction      = passLineList;
			submitLineStripFunction     = passLineStrip;
			submitTriangleListFunction  = passTriangleList;
			submitTriangleStripFunction = passTriangleStrip;
			submitTriangleFanFunction   = passTriangleFan;
		}

	if (primitiveProcessing == PP_CheckMaterialAndVertices)
		if (materialTranslucent)
		{
			submitPointListFunction     = queuePointList;
			submitLineListFunction      = queueLineList;
			submitLineStripFunction     = queueLineStrip;
			submitTriangleListFunction  = queueTriangleList;
			submitTriangleStripFunction = queueTriangleStrip;
			submitTriangleFanFunction   = queueTriangleFan;

			// -qq- OPTIMIZE should not have to setup the material in this case
		}
		else
		{
			submitPointListFunction     = testPointList;
			submitLineListFunction      = testLineList;
			submitLineStripFunction     = testLineStrip;
			submitTriangleListFunction  = testTriangleList;
			submitTriangleStripFunction = testTriangleStrip;
			submitTriangleFanFunction   = testTriangleFan;
	}

	materialData->setup(device);

	// get the material colors
	matR = materialData->getR();
	matG = materialData->getG();
	matB = materialData->getB();
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawPointList(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 1, ("GL_d3d_5::drawPointListU %d vertices", numberOfVertices));
	Geometry::drawPointList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawLineList(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 2 || (numberOfVertices & 1), ("GL_d3d_5::drawLineListU %d vertices", numberOfVertices));
	Geometry::drawLineList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawLineStrip(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 2, ("GL_d3d_5::drawLineStripU %d vertices", numberOfVertices));
	Geometry::drawLineStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawTriangleList(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 3 || (numberOfVertices % 3 != 0), ("GL_d3d_5::drawTriangleListU %d vertices", numberOfVertices));
	Geometry::drawTriangleList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawTriangleStrip(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 3, ("GL_d3d_5::drawTriangleStripU %d vertices", numberOfVertices));
	Geometry::drawTriangleStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawTriangleFan(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 3, ("GL_d3d_5::drawTriangleFan U %d vertices", numberOfVertices));
	Geometry::drawTriangleFan(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawPointList(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 1, ("GL_d3d_5::drawPointListT %d vertices", numberOfVertices));
	submitPointList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawLineList(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 2 || (numberOfVertices & 1), ("GL_d3d_5::drawLineListT %d vertices", numberOfVertices));
	submitLineList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawLineStrip(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 2, ("GL_d3d_5::drawLineStripT %d vertices", numberOfVertices));
	submitLineStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawTriangleList(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 3 || (numberOfVertices % 3 != 0), ("GL_d3d_5::drawTriangleListT %d vertices", numberOfVertices));
	submitTriangleList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawTriangleStrip(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 3, ("GL_d3d_5::drawTriangleStripT %d vertices", numberOfVertices));
	submitTriangleStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::drawTriangleFan(const TVertex *vertexList, int numberOfVertices)
{
	// -qq- need 2d clipping
	DEBUG_FATAL(numberOfVertices < 3, ("GL_d3d_5::drawTriangleFanT %d vertices", numberOfVertices));
	submitTriangleFan(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

void GL_d3d_5::passPointList(const TVertex *vertexList, int numberOfVertices)
{
	D3DTLVERTEX  *v;
	int           i, r, g, b, a;
	HRESULT       result;

	DEBUG_FATAL(numberOfVertices < 1, ("GL_d3d_5::passPointList %d vertices", numberOfVertices));

	// -qq- queue up translucent primitives

	// convert all the vertices to Direct3d vertices
	for (i = numberOfVertices, v = vertexArray; i; --i, ++v, ++vertexList)
	{
		r = REAL_TO_INT(vertexList->r * matR);
		g = REAL_TO_INT(vertexList->g * matG);
		b = REAL_TO_INT(vertexList->b * matB);
		a = REAL_TO_INT(vertexList->a) * 255;
		DEBUG_FATAL(r < 0 || g < 0 || b < 0 || r > 255 || g > 255 || b > 255, ("GL_d3d_5::passPointList bad light"));

		v->dvSX       = D3DVAL(vertexList->x);
		v->dvSY       = D3DVAL(vertexList->y);
		v->dvSZ       = D3DVAL(1-vertexList->ooz);
		v->dvRHW      = D3DVAL(vertexList->ooz);
		v->dcColor    = RGBA_MAKE(r,g,b,a);
		v->dcSpecular = 0;
		v->dvTU       = D3DVAL(vertexList->u);
		v->dvTV       = D3DVAL(vertexList->v);
	}

	do
	{
		result = device->DrawPrimitive(D3DPT_POINTLIST, D3DVT_TLVERTEX, vertexArray, numberOfVertices, 0);
	} while (result == DDERR_WASSTILLDRAWING);

	FATAL_HR("gl_d3d_5::passPointList %d", result);
}

// ----------------------------------------------------------------------

void GL_d3d_5::passLineList(const TVertex *vertexList, int numberOfVertices)
{
	D3DTLVERTEX  *v;
	int           i, r, g, b, a;
	HRESULT       result;

	DEBUG_FATAL(numberOfVertices < 2 || (numberOfVertices & 1), ("GL_d3d_5::passLineList %d vertices", numberOfVertices));

	// -qq- queue up translucent primitives

	// convert all the vertices to Direct3d vertices
	for (i = numberOfVertices, v = vertexArray; i; --i, ++v, ++vertexList)
	{
		r = REAL_TO_INT(vertexList->r * matR);
		g = REAL_TO_INT(vertexList->g * matG);
		b = REAL_TO_INT(vertexList->b * matB);
		a = REAL_TO_INT(vertexList->a) * 255;
		DEBUG_FATAL(r < 0 || g < 0 || b < 0 || r > 255 || g > 255 || b > 255, ("GL_d3d_5::passPointList bad light"));

		v->dvSX       = D3DVAL(vertexList->x);
		v->dvSY       = D3DVAL(vertexList->y);
		v->dvSZ       = D3DVAL(1-vertexList->ooz);
		v->dvRHW      = D3DVAL(vertexList->ooz);
		v->dcColor    = RGBA_MAKE(r,g,b,a);
		v->dcSpecular = 0;
		v->dvTU       = D3DVAL(vertexList->u);
		v->dvTV       = D3DVAL(vertexList->v);
	}

	do
	{
		result = device->DrawPrimitive(D3DPT_LINELIST, D3DVT_TLVERTEX, vertexArray, numberOfVertices, 0);
	} while (result == DDERR_WASSTILLDRAWING);

	FATAL_HR("gl_d3d_5::passLineList %d", result);
}

// ----------------------------------------------------------------------

void GL_d3d_5::passLineStrip(const TVertex *vertexList, int numberOfVertices)
{
	D3DTLVERTEX  *v;
	int           i, r, g, b, a;
	HRESULT       result;

	DEBUG_FATAL(numberOfVertices < 2, ("GL_d3d_5::passLineStrip %d vertices", numberOfVertices));

	// -qq- queue up translucent primitives

	// convert all the vertices to Direct3d vertices
	for (i = numberOfVertices, v = vertexArray; i; --i, ++v, ++vertexList)
	{
		r = REAL_TO_INT(vertexList->r * matR);
		g = REAL_TO_INT(vertexList->g * matG);
		b = REAL_TO_INT(vertexList->b * matB);
		a = REAL_TO_INT(vertexList->a) * 255;
		DEBUG_FATAL(r < 0 || g < 0 || b < 0 || r > 255 || g > 255 || b > 255, ("GL_d3d_5::passPointList bad light"));

		v->dvSX       = D3DVAL(vertexList->x);
		v->dvSY       = D3DVAL(vertexList->y);
		v->dvSZ       = D3DVAL(1-vertexList->ooz);
		v->dvRHW      = D3DVAL(vertexList->ooz);
		v->dcColor    = RGBA_MAKE(r,g,b,a);
		v->dcSpecular = 0;
		v->dvTU       = D3DVAL(vertexList->u);
		v->dvTV       = D3DVAL(vertexList->v);
	}

	do
	{
		result = device->DrawPrimitive(D3DPT_LINESTRIP, D3DVT_TLVERTEX, vertexArray, numberOfVertices, 0);
	} while (result == DDERR_WASSTILLDRAWING);

	FATAL_HR("gl_d3d_5::passLineSrip %d", result);
}

// ----------------------------------------------------------------------

void GL_d3d_5::passTriangleList(const TVertex *vertexList, int numberOfVertices)
{
	D3DTLVERTEX  *v;
	int           i, r, g, b, a;
	HRESULT       result;

	DEBUG_FATAL(numberOfVertices < 3 || (numberOfVertices % 3), ("GL_d3d_5::passTriangleList %d vertices", numberOfVertices));

	// -qq- queue up translucent primitives

	// convert all the vertices to Direct3d vertices
	for (i = numberOfVertices, v = vertexArray; i; --i, ++v, ++vertexList)
	{
		r = REAL_TO_INT(vertexList->r * matR);
		g = REAL_TO_INT(vertexList->g * matG);
		b = REAL_TO_INT(vertexList->b * matB);
		a = REAL_TO_INT(vertexList->a) * 255;
		DEBUG_FATAL(r < 0 || g < 0 || b < 0 || r > 255 || g > 255 || b > 255, ("GL_d3d_5::passPointList bad light"));

		v->dvSX       = D3DVAL(vertexList->x);
		v->dvSY       = D3DVAL(vertexList->y);
		v->dvSZ       = D3DVAL(1-vertexList->ooz);
		v->dvRHW      = D3DVAL(vertexList->ooz);
		v->dcColor    = RGBA_MAKE(r,g,b,a);
		v->dcSpecular = 0;
		v->dvTU       = D3DVAL(vertexList->u);
		v->dvTV       = D3DVAL(vertexList->v);
	}

	do
	{
		result = device->DrawPrimitive(D3DPT_TRIANGLELIST, D3DVT_TLVERTEX, vertexArray, numberOfVertices, 0);
	} while (result == DDERR_WASSTILLDRAWING);

	FATAL_HR("gl_d3d_5::passTriangleList %d", result);
}

// ----------------------------------------------------------------------

void GL_d3d_5::passTriangleStrip(const TVertex *vertexList, int numberOfVertices)
{
	D3DTLVERTEX  *v;
	int           i, r, g, b, a;
	HRESULT       result;

	DEBUG_FATAL(numberOfVertices < 3, ("GL_d3d_5::passTriangleStrip %d vertices", numberOfVertices));

	// -qq- queue up translucent primitives

	// convert all the vertices to Direct3d vertices
	for (i = numberOfVertices, v = vertexArray; i; --i, ++v, ++vertexList)
	{
		r = REAL_TO_INT(vertexList->r * matR);
		g = REAL_TO_INT(vertexList->g * matG);
		b = REAL_TO_INT(vertexList->b * matB);
		a = REAL_TO_INT(vertexList->a) * 255;
		DEBUG_FATAL(r < 0 || g < 0 || b < 0 || r > 255 || g > 255 || b > 255, ("GL_d3d_5::passPointList bad light"));

		v->dvSX       = D3DVAL(vertexList->x);
		v->dvSY       = D3DVAL(vertexList->y);
		v->dvSZ       = D3DVAL(1-vertexList->ooz);
		v->dvRHW      = D3DVAL(vertexList->ooz);
		v->dcColor    = RGBA_MAKE(r,g,b,a);
		v->dcSpecular = 0;
		v->dvTU       = D3DVAL(vertexList->u);
		v->dvTV       = D3DVAL(vertexList->v);
	}

	do
	{
		result = device->DrawPrimitive(D3DPT_TRIANGLESTRIP, D3DVT_TLVERTEX, vertexArray, numberOfVertices, 0);
	} while (result == DDERR_WASSTILLDRAWING);

	FATAL_HR("gl_d3d_5::passTriangleStrip %d", result);
}

// ----------------------------------------------------------------------

void GL_d3d_5::passTriangleFan(const TVertex *vertexList, int numberOfVertices)
{
	D3DTLVERTEX  *v;
	int           i, r, g, b, a;
	HRESULT       result;

	DEBUG_FATAL(numberOfVertices < 3, ("GL_d3d_5::passTriangleFan %d vertices", numberOfVertices));

	// -qq- queue up translucent primitives

	// convert all the vertices to Direct3d vertices
	for (i = numberOfVertices, v = vertexArray; i; --i, ++v, ++vertexList)
	{
		r = REAL_TO_INT(vertexList->r * matR);
		g = REAL_TO_INT(vertexList->g * matG);
		b = REAL_TO_INT(vertexList->b * matB);
		a = REAL_TO_INT(vertexList->a) * 255;
		DEBUG_FATAL(r < 0 || g < 0 || b < 0 || r > 255 || g > 255 || b > 255, ("GL_d3d_5::passPointList bad light"));

		v->dvSX       = D3DVAL(vertexList->x);
		v->dvSY       = D3DVAL(vertexList->y);
		v->dvSZ       = D3DVAL(1-vertexList->ooz);
		v->dvRHW      = D3DVAL(vertexList->ooz);
		v->dcColor    = RGBA_MAKE(r,g,b,a);
		v->dcSpecular = 0;
		v->dvTU       = D3DVAL(vertexList->u);
		v->dvTV       = D3DVAL(vertexList->v);
	}

	do
	{
		result = device->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DVT_TLVERTEX, vertexArray, numberOfVertices, 0);
	} while (result == DDERR_WASSTILLDRAWING);

	FATAL_HR("gl_d3d_5::passTriangleFan %d", result);
}

// ----------------------------------------------------------------------

void GL_d3d_5::testPointList(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 1, ("GL_d3d_5::testPointList %d vertices", numberOfVertices));

	// test the point list
	for ( ; numberOfVertices; --numberOfVertices, ++vertexList)
		if (vertexList->a != CONST_REAL(1))
			TranslucentQueue::addPoint(vertexList);
		else
			passPointList(vertexList, 1);
}

// ----------------------------------------------------------------------

void GL_d3d_5::testLineList(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 2 || (numberOfVertices & 1), ("GL_d3d_5::testLineList %d vertices", numberOfVertices));

	// test the line list
	for ( ; numberOfVertices; vertexList += 2, numberOfVertices -= 2)
		if (vertexList[0].a != CONST_REAL(1) || vertexList[1].a != CONST_REAL(1))
			TranslucentQueue::addLine(vertexList + 0, vertexList + 1);
		else
			passLineList(vertexList, 2);
}

// ----------------------------------------------------------------------

void GL_d3d_5::testLineStrip(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 2, ("GL_d3d_5::testLineStrip %d vertices", numberOfVertices));

	// test the line strip
	for (numberOfVertices--; numberOfVertices; ++vertexList, --numberOfVertices)
		if (vertexList[0].a != CONST_REAL(1) || vertexList[1].a != CONST_REAL(1))
			TranslucentQueue::addLine(vertexList + 0, vertexList + 1);
		else
			passLineList(vertexList, 2);
}

// ----------------------------------------------------------------------

void GL_d3d_5::testTriangleList(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 3 || (numberOfVertices % 3 != 0), ("GL_d3d_5::testTriangleList %d vertices", numberOfVertices));

	// test the triangle list
	for ( ; numberOfVertices; vertexList += 3, numberOfVertices -= 3)
		if (vertexList[0].a != CONST_REAL(1) || vertexList[1].a != CONST_REAL(1) || vertexList[2].a != CONST_REAL(1))
		{
			DPRINT(("%f %f %f\n", vertexList[0].a, vertexList[1].a, vertexList[2].a));
			TranslucentQueue::addTriangle(vertexList + 0, vertexList + 1, vertexList + 2);
		}
		else
			passTriangleList(vertexList, 3);
}

// ----------------------------------------------------------------------

void GL_d3d_5::testTriangleStrip(const TVertex *vertexList, int numberOfVertices)
{
	bool    odd;
	TVertex temp[3];

	DEBUG_FATAL(numberOfVertices < 3, ("GL_d3d_5::testTriangleStrip %d vertices", numberOfVertices));

	// queue the triangle strip
	for (odd = false, numberOfVertices -= 2 ; numberOfVertices; ++vertexList, --numberOfVertices, odd = !odd)
		if (vertexList[0].a != CONST_REAL(1) || vertexList[1].a != CONST_REAL(1) || vertexList[2].a != CONST_REAL(1))
		{
			if (odd)
				TranslucentQueue::addTriangle(vertexList + 0, vertexList + 2, vertexList + 1);
			else
				TranslucentQueue::addTriangle(vertexList + 0, vertexList + 1, vertexList + 2);
		}
		else
		{
			if (odd)
			{
				// copy the vertices into a clockwise temporary buffer
				temp[0] = vertexList[0];
				temp[1] = vertexList[2];
				temp[2] = vertexList[1];

				passTriangleList(temp, 3);
			}
			else
				passTriangleList(vertexList, 3);
		}
}

// ----------------------------------------------------------------------

void GL_d3d_5::testTriangleFan(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 3, ("GL_d3d_5::testTriangleFan %d vertices", numberOfVertices));

	const TVertex *v0 = vertexList;

	TVertex temp[3];

	// copy the first vertex one time
	temp[0] = *v0;

	for (numberOfVertices -= 2, ++vertexList; numberOfVertices; ++vertexList, --numberOfVertices)
		if (v0->a != CONST_REAL(1) || vertexList[0].a != CONST_REAL(1) || vertexList[1].a != CONST_REAL(1))
			TranslucentQueue::addTriangle(v0, vertexList + 0, vertexList + 1);
		else
		{
			// copy the rest of the triangle into a temporary buffer
			temp[1] = vertexList[0];
			temp[2] = vertexList[1];

			passTriangleList(temp, 3);
		}
}

// ----------------------------------------------------------------------

void GL_d3d_5::queuePointList(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 1, ("GL_d3d_5::queuePointList %d vertices", numberOfVertices));

	// queue the point list
	for ( ; numberOfVertices; --numberOfVertices, ++vertexList)
		TranslucentQueue::addPoint(vertexList);
}

// ----------------------------------------------------------------------

void GL_d3d_5::queueLineList(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 2 || (numberOfVertices & 1), ("GL_d3d_5::queueLineList %d vertices", numberOfVertices));

	// queue the line list
	for ( ; numberOfVertices; vertexList += 2, numberOfVertices -= 2)
		TranslucentQueue::addLine(vertexList, vertexList + 1);
}

// ----------------------------------------------------------------------

void GL_d3d_5::queueLineStrip(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 2, ("GL_d3d_5::queueLineStrip %d vertices", numberOfVertices));

	// queue the line strip
	for (numberOfVertices--; numberOfVertices; ++vertexList, --numberOfVertices)
		TranslucentQueue::addLine(vertexList, vertexList + 1);
}

// ----------------------------------------------------------------------

void GL_d3d_5::queueTriangleList(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 3 || (numberOfVertices % 3 != 0), ("GL_d3d_5::queueTriangleList %d vertices", numberOfVertices));

	// queue the triangle list
	for ( ; numberOfVertices; vertexList += 3, numberOfVertices -= 3)
		TranslucentQueue::addTriangle(vertexList + 0, vertexList + 1, vertexList + 2);
}

// ----------------------------------------------------------------------
// -qq- optimize by unrolling the loop to get rid of the if test

void GL_d3d_5::queueTriangleStrip(const TVertex *vertexList, int numberOfVertices)
{
	bool odd;

	DEBUG_FATAL(numberOfVertices < 3, ("GL_d3d_5::queueTriangleStrip %d vertices", numberOfVertices));

	// queue the triangle strip
	for (odd = false, numberOfVertices -= 2 ; numberOfVertices; ++vertexList, --numberOfVertices, odd = !odd)
		if (odd)
			TranslucentQueue::addTriangle(vertexList + 0, vertexList + 2, vertexList + 1);
		else
			TranslucentQueue::addTriangle(vertexList + 0, vertexList + 1, vertexList + 2);
}

// ----------------------------------------------------------------------

void GL_d3d_5::queueTriangleFan(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(numberOfVertices < 3, ("GL_d3d_5::queueTriangleFan %d vertices", numberOfVertices));

	// queue the triangle strip
	const TVertex *v0 = vertexList;
	for (numberOfVertices -= 2, ++vertexList; numberOfVertices; ++vertexList, --numberOfVertices)
		TranslucentQueue::addTriangle(v0, vertexList + 0, vertexList + 1);
}
// ----------------------------------------------------------------------

void GL_d3d_5::movie(const uv_instance *uv)
{
#if 1
	UNREF(uv);
#else

	u_int32 squareXOffset,squareYOffset,ssYoffset,ssXoffset,adjWidth,adjHeight;
	D3DTLVERTEX           *v;
	PolyVertex             pv[4];
	int                    result,ndx,i;

	// shut off alpha for these polys
	result = device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
	FATAL_HR("gl_d3d_5::movie alphaBlend\n", result);

	adjWidth            = uv->width;
	adjHeight           = uv->height;

	if (uv->options.option_flags & (X_DOUBLE | X_SKIP))
		adjWidth*=2;
	if (uv->options.option_flags & (Y_DOUBLE | Y_SKIP))
		adjHeight*=2;
	if (uv->options.option_flags & X_CENTRE)
		if (uv->options.option_flags & (X_DOUBLE | X_SKIP))
			ssXoffset = (width - adjWidth) / 2;
		else
			ssXoffset = (width - adjWidth) / 2;
	else
		if (uv->options.option_flags & X_OFFSET)
			ssXoffset = uv->options.x_offset;
		else
			ssXoffset = 0;

	if (uv->options.option_flags & Y_CENTRE)
		if (uv->options.option_flags & (Y_DOUBLE | Y_SKIP))
			ssYoffset = (height - adjHeight) / 2;
		else
			ssYoffset = (height - adjHeight) / 2;
	else
		if (uv->options.option_flags & Y_OFFSET)
			ssYoffset = uv->options.y_offset;
		else
			ssYoffset = 0;

	// walk by Y in squares of 256x256 (in destination screen space)
	for (squareYOffset=0; squareYOffset<adjHeight; squareYOffset+=256)  
	{
		// walk by X in squares of 256x256 (in destination screen space)
		for (squareXOffset=0; squareXOffset<adjWidth; squareXOffset+=256)
		{
			// don't extend beyond the edges of the movie source
			int cappedX=min(squareXOffset+256,adjWidth);
			int cappedY=min(squareYOffset+256,adjHeight);

			pv[0].x = float(squareXOffset+ssXoffset);
			pv[0].y = float(squareYOffset+ssYoffset);
			pv[0].u = float(0);
			pv[0].v = float(0);
			pv[1].x = float(cappedX      +ssXoffset);
			pv[1].y = float(squareYOffset+ssYoffset);
			pv[1].u = float(cappedX-squareXOffset) / 256.0f;
			pv[1].v = float(0);
			pv[2].x = float(cappedX+ssXoffset);
			pv[2].y = float(cappedY+ssYoffset);
			pv[2].u = float(cappedX-squareXOffset) / 256.0f;
			pv[2].v = float(cappedY-squareYOffset) / 256.0f;
			pv[3].x = float(squareXOffset+ssXoffset);
			pv[3].y = float(cappedY      +ssYoffset);
			pv[3].u = float(0);
			pv[3].v = float(cappedY-squareYOffset)  / 256.0f;

			// convert all the vertices to Direct3d vertices
			for (ndx=0, i = 4, v = vertexArray; i; i--, v++, ndx++)
			{
				v->dvSX       = D3DVAL(pv[ndx].x);
				v->dvSY       = D3DVAL(pv[ndx].y);
				v->dvSZ       = D3DVAL(0);
				v->dvRHW      = D3DVAL(1);
				v->dcColor    = RGBA_MAKE(255,255,255,255);
				v->dcSpecular = 0;
				v->dvTU       = D3DVAL(pv[ndx].u);
				v->dvTV       = D3DVAL(pv[ndx].v);
			}

			MaterialData_d3d_5  *movieSquare = NEW MaterialData_d3d_5(uv, squareXOffset, squareYOffset);

			// handle material change
			result = device->SetRenderState(D3DRENDERSTATE_TEXTUREHANDLE, movieSquare->getTextureHandle(device));
			FATAL_HR("gl_d3d_5::movie txhandle %d\n", result);

			do
			{
				result = device->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DVT_TLVERTEX, vertexArray, 4, 0);
			} while (result == DDERR_WASSTILLDRAWING);
			FATAL_HR("gl_d3d_5::movie %d", result);
		}
	}
#endif
}

// ----------------------------------------------------------------------

bool GL_d3d_5::specialEffect(int id, void *data)
{
	return GL_dd::specialEffect(id, data);
}

// ======================================================================